@@ -24,6 +24,9 @@
 #include <pango/pango-ot.h>
 #include "pango-ot-private.h"
 
+#define PANGO_SCALE_26_6 (PANGO_SCALE / (1<<6))
+#define PANGO_UNITS_26_6(d) (PANGO_SCALE_26_6 * (d))
+
 typedef struct _PangoOTRule PangoOTRule;
 
 struct _PangoOTRule 
@@ -233,14 +236,51 @@ pango_ot_ruleset_shape (PangoOTRuleset   *ruleset,
     }
   else
     result_string = in_string;
-  
+
+  if (gpos)
+    {
+      TTO_GPOS_Data *outgpos = NULL;
+
+      if (!TT_GPOS_Apply_String (ruleset->info->face, gpos, 0, result_string, &outgpos,
+				 FALSE /* enable device-dependant values */,
+				 FALSE /* Even though this might be r2l text, RTL is handled elsewhere */))
+	{
+	  for (i = 0; i < result_string->length; i++)
+	    {
+	      FT_Pos x_pos = outgpos[i].x_pos;
+	      FT_Pos y_pos = outgpos[i].y_pos;
+	      int back = i;
+	      int j;
+
+	      while (outgpos[back].back != 0)
+		{
+		  back  -= outgpos[back].back;
+		  x_pos += outgpos[back].x_pos;
+		  y_pos += outgpos[back].y_pos;
+		}
+
+	      for (j = back; j < i; j++)
+	        glyphs->glyphs[i].geometry.x_offset -= glyphs->glyphs[j].geometry.width;
+
+	      glyphs->glyphs[i].geometry.x_offset += PANGO_UNITS_26_6(x_pos);
+	      glyphs->glyphs[i].geometry.y_offset += PANGO_UNITS_26_6(y_pos);
+
+	      if (outgpos[i].new_advance)
+		glyphs->glyphs[i].geometry.width  = PANGO_UNITS_26_6(outgpos[i].x_advance);
+	      else
+		glyphs->glyphs[i].geometry.width += PANGO_UNITS_26_6(outgpos[i].x_advance);
+	    }
+
+	  FT_Free(gpos->memory, (void *)outgpos);
+	}
+    }
+
   pango_glyph_string_set_size (glyphs, result_string->length);
 
   last_cluster = -1;
   for (i = 0; i < result_string->length; i++)
     {
       glyphs->glyphs[i].glyph = result_string->string[i];
-      glyphs->glyphs[i].glyph = result_string->string[i];
 
       glyphs->log_clusters[i] = result_string->logClusters[i];
       if (glyphs->log_clusters[i] != last_cluster)
